### saddlepoint approx: modified from Lumley survey package.
saddle = function(x,lambda){
  d = max(lambda)
  lambda = lambda/d
  x = x/d
  k0 = function(zeta) -sum(log(1-2*zeta*lambda))/2
  kprime0 = function(zeta) sapply(zeta, function(zz) sum(lambda/(1-2*zz*lambda)))
  kpprime0 = function(zeta) 2*sum(lambda^2/(1-2*zeta*lambda)^2)
  n = length(lambda)
  if (any(lambda < 0)) {
    lmin = max(1/(2 * lambda[lambda < 0])) * 0.99999
  } else if (x>sum(lambda)){
    lmin = -0.01
  } else {
    lmin = -length(lambda)/(2*x)
  }
  lmax = min(1/(2*lambda[lambda>0]))*0.99999
  hatzeta = uniroot(function(zeta) kprime0(zeta) - x, lower = lmin, upper = lmax, tol = 1e-08)$root
  w = sign(hatzeta)*sqrt(2*(hatzeta*x-k0(hatzeta)))
  v = hatzeta*sqrt(kpprime0(hatzeta))
  if(abs(hatzeta)<1e-4){
    return(NA)
  } else{
    return( pnorm(w+log(v/w)/w, lower.tail=FALSE) )
  }
}
Sadd.pval = function(Q.all,lambda){
  sad = rep(1,length(Q.all))
  if(sum(Q.all>0)>0){
    sad[Q.all>0] = sapply(Q.all[Q.all>0],saddle,lambda=lambda)
  }
  id = which(is.na(sad))
  if(length(id)>0){
    sad[id] = Liu.pval(Q.all[id], lambda)
  }
  return(sad)
}
### modified Liu method from Lee SKAT-O paper
Liu.pval = function(Q, lambda){
  c1 = rep(0,4); for(i in 1:4){ c1[i] = sum(lambda^i) }
  muQ = c1[1];  sigmaQ = sqrt(2 *c1[2])
  s1 = c1[3]/c1[2]^(3/2);  s2 = c1[4]/c1[2]^2
  if(s1^2 > s2){
    a = 1/(s1 - sqrt(s1^2 - s2));  d = s1 *a^3 - a^2;  l = a^2 - 2*d
  } else {
    l = 1/s2;  a = sqrt(l);  d = 0
  }
  muX = l+d;  sigmaX = sqrt(2)*a
  
  Q.Norm = (Q - muQ)/sigmaQ
  Q.Norm1 = Q.Norm*sigmaX + muX
  pchisq(Q.Norm1, df = l,ncp=d, lower.tail=FALSE)
}
## Compute the tail probability of 1-DF chi-square mixtures
##
## Use the Davies method (from CompQuadForm R package); if fail, switch to the saddlepoint approximation following Wu et al. (2016).
## Internal function. Not to be called directly.
## We adopt a stringent convergence criterion for the Davies method to avoid aberrant results (Wu et.al, 2016).
## @param Q.all  test statistics
## @param lambda  mixing coefficients
## @param acc  error bound
## @param lim  maximum number of integration terms
## @references
## Wu,B., Guan,W., Pankow,J.S. (2016) On efficient and accurate calculation of significance p-values for sequence kernel association test of variant set. \emph{Annals of human genetics}, 80(2), 123-135.
KAT.pval <- function(Q.all, lambda, acc=1e-9,lim=1e6){
  pval = rep(0, length(Q.all))
  i1 = which(is.finite(Q.all))
  for(i in i1){
    tmp = davies(Q.all[i],lambda,acc=acc,lim=lim); pval[i] = tmp$Qq
    if((tmp$ifault>0)|(pval[i]<=0)|(pval[i]>=1)) pval[i] = Sadd.pval(Q.all[i],lambda)
  }
  return(pval)
}

#' @importFrom CompQuadForm davies
NULL




#' Fit a null multi-trait regression model
#'
#' Compute a null multivariate regression model
#' @param Ys  matrix of continuous outcomes with variables in column
#' @param X   matrix of covariates to be adjusted
#' @export
MSKAT.cnull <- function(Ys,X=NULL){
  if(class(Ys)!='matrix') Ys = as.matrix(Ys)
  N = dim(Ys)[1]
  ## null model fitting
  if(is.null(X)){
    px = 1
    Ux = matrix(1/sqrt(N), N,1)
  } else{
    if(class(X)!='matrix') X = as.matrix(X)
    px = dim(X)[2]+1
    Ux = svd(cbind(1,X), nv=0)$u
  }
  U0 = Ys - Ux%*%(t(Ux)%*%Ys)
  Sig = cor(U0); iSig = solve(Sig)
  a1 = svd(Sig,nv=0); ihSig = a1$u%*%diag(1/sqrt(a1$d))%*%t(a1$u)
  V0 = colSums(U0^2)/(N-px)
  return(list(Ux=Ux,U0=U0,V0=V0,Sig=Sig,iSig=iSig,ihSig=ihSig,svd=a1, Ys=Ys,px=px) )
}

#' Multi-trait sequence kernel association test (SKAT) for rare variant set association
#'
#' Compute the rare variant set association test statistics and p-values
#' @param obj  fitted null model from MSKAT.cnull
#' @param G    genotype matrix with SNPs in column
#' @param W    variant weight; use Beta dist weight if NULL
#' @param W.beta   Beta distribution parameters for the variant weight
#' @return a p-value vector p.value with two components (see Wu and Pankow, 2016)
#' \describe{
#'   \item{Q}{ p-value for the kernel statistic Q (sum of chi-square statstics across variants) }
#'   \item{Q'}{ p-value for the kernel statistic Q' (sum of SKAT statistics across traits) }
#' }
#' @export
#' @references
#' Wu,B. and Pankow,J.S. (2016) Sequence kernel association test of multiple continuous phenotypes. \emph{Genetic Epidemiology}, 40(2), 91-100.
#' @examples
#' library(CompQuadForm)
#' n=2e3; K=4; p = 2; m = 50
#' X = matrix(rnorm(n*p), n,p); Z = rbinom(n,1,0.5); XZ = cbind(X,Z)
#' ## simulate genotype from MVN
#' rho = 0.25; maf1 = 0.005; maf2 = 0.015
#' q1 = qnorm(maf1); q2 = qnorm(maf2)
#' ## pairwise LD R^2 calc with MVN: 0.023^2 and 0.042^2 (very low)
#' ## library(mvtnorm)
#' ## sig = matrix(rho,2,2); diag(sig) = 1
#' ## (pmvnorm(upper=c(q1,q1),sigma=sig)-maf1^2)/maf1/(1-maf1)
#' ## (pmvnorm(upper=c(q2,q2),sigma=sig)-maf2^2)/maf2/(1-maf2)
#' ## sim
#' qz = q1 + Z*(q2-q1)
#' G1 = matrix(rnorm(n*m), n,m)*sqrt(1-rho) + rnorm(n)*sqrt(rho)
#' G2 = matrix(rnorm(n*m), n,m)*sqrt(1-rho) + rnorm(n)*sqrt(rho)
#' G = 1*(G1<qz) + 1*(G2<qz)
#' yrho = 0.25
#' Y = matrix(rnorm(n*K), n,K)*sqrt(1-yrho) + rnorm(n)*sqrt(yrho)
#' Y[,1] = Y[,1] + rowSums(X) + Z - rowSums(G[,1:3])*0.5 + G[,m]*0.7
#' Y[,2] = Y[,2] + rowSums(X) + Z + rowSums(G[,c(1,4,5)])*0.5 - G[,m-1]*0.6
#' MSKAT(MSKAT.cnull(Y,XZ), G, W.beta=c(1,25))
#' gamut(Y,G,XZ)
MSKAT <- function(obj, G, W=NULL, W.beta=c(1,25)){
  if(class(G)!='matrix') G = as.matrix(G)
  n = dim(G)[1];  m = dim(G)[2]; K = dim(obj$Ys)[2]
  if(is.null(W)){
    W = dbeta(colMeans(G)/2, W.beta[1],W.beta[2])
    W = W/sum(W)*m
  } else{
    if(length(W)<=1) W = rep(1,m)
  }
  ## KAT statistic
  Ux = obj$Ux; U0 = obj$U0; V0 = obj$V0
  S = t(t(U0)%*%G/sqrt(V0)) ## m,K
  Ge = G - Ux%*%(t(Ux)%*%G); Ge2 = t(Ge)%*%Ge
  R = t(Ge2*W)*W
  Lam = zapsmall( svd(R, nu=0,nv=0)$d )
  Lam = Lam[Lam>0]
  ## Q
  ihSig = obj$ihSig
  Z = W*S%*%ihSig
  Q = sum(Z^2)
  pval = KAT.pval(Q,rep(Lam,K))
  ## Q'
  Z = W*S
  Lam1 = svd(obj$Sig,nu=0,nv=0)$d
  Qp = sum(Z^2)
  pvalp = KAT.pval(Qp, c(outer(Lam,Lam1)))
  p.value=c(pval, pvalp)
  names(p.value) = c("Q", "Q'")
  return( p.value )
}

